-- Greg Stitt
-- University of Florida
--
-- Description: This entity discusses the possibility of extending the sum
-- entity to support a generic number of inputs and a generic width.
-- Unforunately, a limitation of VHDL (pre 2008) makes this nearly impossible.
--
-- Topics discussed: Arrays of unconstrained types

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.sum_pkg.all;                   

-- Ideally, we would like to add a width generic to support both a generic
-- number of inputs and a generic width.

entity sum is
  generic (
    num_inputs : positive := 4;
    width : positive := 8);
  port (
    
    -- Unfortunately, this would require use to change the type definition of
    -- input_array to be:
    -- type input_array is array (natural range <>) of std_logic_vector
    -- Note that this is array defines each element using an unconstrained
    -- type. VHDL (pre 2008) does not support such arrays. Therefore, we have
    -- to make the width constant.  
    
    inputs     : input_array(0 to num_inputs-1);

    -- Fortunately, we can make things a little better by using a constant
    -- inside of sum_pkg instead of hardcoding everything to a specific width.
    -- When I defined a constant in a package, I use a C_ prefix so I know
    -- where the constant comes from.
    
    output     : out std_logic_vector(C_SUM_WIDTH-1 downto 0));
end sum;

architecture BHV of sum is
begin  -- BHV

  process(inputs)
    variable temp_sum : unsigned(C_SUM_WIDTH-1 downto 0);    
  begin

    temp_sum := to_unsigned(0, temp_sum'length);

    for i in 0 to num_inputs-1 loop
      temp_sum := temp_sum + unsigned(inputs(i));      
    end loop;

    output <= std_logic_vector(temp_sum);
    
  end process;
end BHV;

-- Alternatively, we could have supported generic widths and numbers of inputs
-- by redefining the type of input. e.g.
--
-- inputs : in std_logic_vector(num_inputs*width-1 downto 0)
--
-- This code will define the right number of bits for the inputs, but it will
-- often be very tricky to index into this in the correct way. It can almost
-- always be done, it just is usually very painful.
